// WSPR Beacon CSG+ 21.09.24

#include <si5351.h>
#include <JTEncode.h>
#include <TimeLib.h>
#include <TinyGPS.h>
#include <LiquidCrystal.h>


#define TONE_SPACING 146  // ~1.46 Hz
#define WSPR_DELAY 683    // Delay value for WSPR
#define SYMBOL_COUNT WSPR_SYMBOL_COUNT
#define CORRECTION 86  // Change this for your ref osc
#define TIME_ZONE 3.0
#define SI5351_REF 25000000UL  // Change this to the frequency of the crystal


#define MAXCH 7  // Change this according to the number of bands inserted below
unsigned long freq0[MAXCH] = { 7040100UL, 10140200UL, 14097100UL, 18106100UL, 21096100UL, 24926100UL, 28126100UL };

char call[7] = "RA4FHE";  // Call
char loc[5] = "LO23";     // Locator
uint8_t dbm = 10;         // Min = 0 dBm, Max = 43 dBm, steps 0, 3, 7, 10, 13, 17, 20, 23, 27, 30, 33, 37, 40, 43
uint8_t tx_buffer[SYMBOL_COUNT];
uint8_t sec = 0;
uint8_t secLlast = 255;
int ch = 0;      // automatic frequency switch index
long randValue = 0;
const int GPS_ON = 6;  // GPIO Power GPS Module ATGM322D
const int rs = 7, en = 8, d4 = 9, d5 = 10, d6 = 11, d7 = 12;


LiquidCrystal lcd(rs, en, d4, d5, d6, d7);
JTEncode jtencode;
Si5351 si5351;
TinyGPS gps;

uint8_t syncGps() {
  uint8_t res = 0;
  while (Serial.available()) {
    if (gps.encode(Serial.read())) {
      unsigned long age;
      int year;
      byte month, day, hour, minute, second;
      gps.crack_datetime(&year, &month, &day, &hour, &minute, &second, NULL, &age);
      if (age < 500) {
        setTime(hour, minute, second, day, month, year);
        adjustTime(TIME_ZONE * SECS_PER_HOUR);
        res = 1;
      }
    }
  }
  return res;
}

void printTx(bool tx) {
  if (tx) {
    lcd.setCursor(9, 1);
    lcd.print("WSPR TX");
  } else {
    lcd.setCursor(9, 1);
    lcd.print("IDLE   ");
  }
}

unsigned long getFreq() {
  return freq0[ch] + randValue;
}

void printFreq() {
  lcd.setCursor(0, 0);
  lcd.print(getFreq());
  lcd.print(" Hz ");
}

void printTime(bool printSync) {
  if (printSync) {
    lcd.setCursor(0, 1);
    lcd.print("Get Time");
  } else {
    lcd.setCursor(0, 1);
    lcd.display();
    if (hour() < 10) lcd.print("0");
    lcd.print(hour());
    lcd.print(":");
    if (minute() < 10) lcd.print("0");
    lcd.print(minute());
    lcd.print(":");
    if (second() < 10) lcd.print("0");
    lcd.print(second());
    lcd.print(" ");
 }
}

// Loop through the string, transmitting one character at a time.
void encode() {
  uint8_t i;
  si5351.set_freq(getFreq() * SI5351_FREQ_MULT, SI5351_CLK2);
  si5351.set_clock_pwr(SI5351_CLK2, 1);
  printFreq();
  uint32_t period = millis() - WSPR_DELAY;
  sec = second();
  printTime(false);
  printTx(true);
  i = 0;
  while (i < SYMBOL_COUNT) {
    uint32_t curMill = millis();
    if (curMill - period >= WSPR_DELAY) {
      period = curMill;
      si5351.set_freq((getFreq() * SI5351_FREQ_MULT) + (tx_buffer[i] * TONE_SPACING), SI5351_CLK2);
      ++i;
    }
    if (sec != second()) {
      sec = second();
      printTime(false);
    }
  }
  // Turn off the output
  si5351.set_clock_pwr(SI5351_CLK2, 0);
  printTx(false);
  //change band for next tx
  ch++;
  if (ch == MAXCH) ch = 0;  // reset band index to start again from first band the in array
  randValue = 50 - random(0, 101);
  printFreq();
}


void setup() {

  Serial.begin(9600);
  lcd.begin(16, 2);
  pinMode(GPS_ON, OUTPUT);
  digitalWrite(GPS_ON, HIGH);
  lcd.setCursor(0, 0);
  lcd.print("WSPR Beacon CSG+");
  lcd.setCursor(5, 1);
  lcd.print("RA4FHE");
  delay(3000);
  lcd.clear();
  printTime(true);
  while (syncGps() == 0) {
    delay(10);
  }
  // Initialize the Si5351
  si5351.init(SI5351_CRYSTAL_LOAD_0PF, SI5351_REF, CORRECTION);
  si5351.drive_strength(SI5351_CLK2, SI5351_DRIVE_8MA);  // Set for max power
  si5351.set_clock_pwr(SI5351_CLK2, 0);                  // Disable the clock initially
  jtencode.wspr_encode(call, loc, dbm, tx_buffer);
}

void loop() {
  sec = second();
  if (minute() % 2 == 0 && sec == 0) {
    encode();
  }
  syncGps();
  if (secLlast != sec) {
    secLlast = sec;
    printTime(false);
  }
}
